@@ -1,6 +1,4 @@
 /**
-2d Cartesian coordinate system.
-
 \file
 \copyright Copyright 2020. Tom de Geus. All rights reserved.
 \license This project is released under the MIT License.
@@ -9,11 +7,11 @@
 #ifndef GMATTENSOR_CARTESIAN2D_H
 #define GMATTENSOR_CARTESIAN2D_H
 
-#include <xtensor/xtensor.hpp>
 #include <xtensor/xadapt.hpp>
 #include <xtensor/xnoalias.hpp>
-#include <xtensor/xview.hpp>
 #include <xtensor/xrandom.hpp>
+#include <xtensor/xtensor.hpp>
+#include <xtensor/xview.hpp>
 
 #include "config.h"
 #include "version.h"
@@ -421,6 +419,42 @@ Same as A4_ddot_B2(const T& A, const U& B) but writes to externally allocated ou
 template <class T, class U, class R>
 inline void A4_ddot_B2(const T& A, const U& B, R& ret);
 
+/**
+Size of the underlying array.
+
+\param A [..., 2, 2] array.
+\return `prod([...])`.
+*/
+template <class T>
+inline auto underlying_size_A2(const T& A);
+
+/**
+Size of the underlying array.
+
+\param A [..., 2, 2] array.
+\return `prod([...])`.
+*/
+template <class T>
+inline auto underlying_size_A4(const T& A);
+
+/**
+Shape of the underlying array.
+
+\param A [..., 2, 2] array.
+\return `[...]`.
+*/
+template <class T>
+inline auto underlying_shape_A2(const T& A);
+
+/**
+Shape of the underlying array.
+
+\param A [..., 2, 2] array.
+\return `[...]`.
+*/
+template <class T>
+inline auto underlying_shape_A4(const T& A);
+
 /**
 Array of tensors:
 -   scalars: shape ``[...]``.
@@ -430,10 +464,8 @@ Array of tensors:
 \tparam N The rank of the array (the actual rank is increased with the tensor-rank).
 */
 template <size_t N>
-class Array
-{
+class Array {
 public:
-
     /**
     Rank of the array (the actual rank is increased with the tensor-rank).
     */
@@ -455,7 +487,6 @@ public:
     */
     std::array<size_t, N> shape() const;
 
-
     /**
     Array of Cartesian2d::O2()
 
@@ -512,8 +543,27 @@ public:
     */
     xt::xtensor<double, N + 4> I4d() const;
 
-protected:
+    /**
+    Return a view of the `i`th second order tensor.
+    For example, consider an array `A` of shape `[M, N, 3, 3]` and you want to obtain the tensor
+    `A[m, n, :, :]` then you should call `view_tensor2(A, m * N + n)`.
 
+    \param var The variable to view (careful not to operate a view that is out-of-scope).
+    \param i The index of the flat storage of the corresponding scalar array.
+    */
+    template <class T>
+    auto view_tensor2(T& var, size_t i) const;
+
+    /**
+    Similar to view_tensor2() for fourth order tensors.
+
+    \param var The variable to view (careful not to operate a view that is out-of-scope).
+    \param i The index of the flat storage of the corresponding scalar array.
+    */
+    template <class T>
+    auto view_tensor4(T& var, size_t i) const;
+
+protected:
     /**
     Constructor 'alias'. Can be used by constructor of derived classes.
 
@@ -554,176 +604,176 @@ No arrays of tensors are allowed, hence the input is fixed to:
 */
 namespace pointer {
 
-    /**
-    See Cartesian2d::O2()
+/**
+See Cartesian2d::O2()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void O2(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void O2(T* ret);
 
-    /**
-    See Cartesian2d::O4()
+/**
+See Cartesian2d::O4()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void O4(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void O4(T* ret);
 
-    /**
-    See Cartesian2d::I2()
+/**
+See Cartesian2d::I2()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void I2(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void I2(T* ret);
 
-    /**
-    See Cartesian2d::II()
+/**
+See Cartesian2d::II()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void II(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void II(T* ret);
 
-    /**
-    See Cartesian2d::I4()
+/**
+See Cartesian2d::I4()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void I4(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void I4(T* ret);
 
-    /**
-    See Cartesian2d::I4rt()
+/**
+See Cartesian2d::I4rt()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void I4rt(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void I4rt(T* ret);
 
-    /**
-    See Cartesian2d::I4s()
+/**
+See Cartesian2d::I4s()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void I4s(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void I4s(T* ret);
 
-    /**
-    See Cartesian2d::I4d()
+/**
+See Cartesian2d::I4d()
 
-    \param ret output 2nd order tensor
-    */
-    template <class T>
-    inline void I4d(T* ret);
+\param ret output 2nd order tensor
+*/
+template <class T>
+inline void I4d(T* ret);
 
-    /**
-    See Cartesian2d::Trace()
+/**
+See Cartesian2d::Trace()
 
-    \param A 2nd order tensor
-    \return scalar
-    */
-    template <class T>
-    inline T Trace(const T* A);
+\param A 2nd order tensor
+\return scalar
+*/
+template <class T>
+inline T Trace(const T* A);
 
-    /**
-    See Cartesian2d::Hydrostatic()
+/**
+See Cartesian2d::Hydrostatic()
 
-    \param A 2nd order tensor
-    \return scalar
-    */
-    template <class T>
-    inline T Hydrostatic(const T* A);
+\param A 2nd order tensor
+\return scalar
+*/
+template <class T>
+inline T Hydrostatic(const T* A);
 
-    /**
-    See Cartesian2d::Sym()
+/**
+See Cartesian2d::Sym()
 
-    \param A 2nd order tensor
-    \param ret 2nd order tensor, may be the same pointer as ``A``
-    */
-    template <class T>
-    inline void sym(const T* A, T* ret);
+\param A 2nd order tensor
+\param ret 2nd order tensor, may be the same pointer as ``A``
+*/
+template <class T>
+inline void sym(const T* A, T* ret);
 
-    /**
-    Returns Cartesian2d::Hydrostatic() and computes Cartesian2d::Deviatoric()
+/**
+Returns Cartesian2d::Hydrostatic() and computes Cartesian2d::Deviatoric()
 
-    \param A 2nd order tensor
-    \param ret 2nd order tensor, may be the same pointer as ``A``
-    \return scalar
-    */
-    template <class T>
-    inline T Hydrostatic_deviatoric(const T* A, T* ret);
+\param A 2nd order tensor
+\param ret 2nd order tensor, may be the same pointer as ``A``
+\return scalar
+*/
+template <class T>
+inline T Hydrostatic_deviatoric(const T* A, T* ret);
 
-    /**
-    Double tensor contraction of the tensor's deviator
+/**
+Double tensor contraction of the tensor's deviator
 
-    \f$ (dev(A))_{ij} (dev(A))_{ji} \f$
+\f$ (dev(A))_{ij} (dev(A))_{ji} \f$
 
-    \param A 2nd order tensor
-    \return scalar
-    */
-    template <class T>
-    inline T Deviatoric_ddot_deviatoric(const T* A);
+\param A 2nd order tensor
+\return scalar
+*/
+template <class T>
+inline T Deviatoric_ddot_deviatoric(const T* A);
 
-    /**
-    See Cartesian2d::Norm_deviatoric()
+/**
+See Cartesian2d::Norm_deviatoric()
 
-    \param A 2nd order tensor
-    \return scalar
-    */
-    template <class T>
-    inline T Norm_deviatoric(const T* A);
+\param A 2nd order tensor
+\return scalar
+*/
+template <class T>
+inline T Norm_deviatoric(const T* A);
 
-    /**
-    See Cartesian2d::A2_ddot_B2()
+/**
+See Cartesian2d::A2_ddot_B2()
 
-    \param A 2nd order tensor
-    \param B 2nd order tensor
-    \return scalar
-    */
-    template <class T>
-    inline T A2_ddot_B2(const T* A, const T* B);
+\param A 2nd order tensor
+\param B 2nd order tensor
+\return scalar
+*/
+template <class T>
+inline T A2_ddot_B2(const T* A, const T* B);
 
-    /**
-    See Cartesian2d::A2s_ddot_B2s()
+/**
+See Cartesian2d::A2s_ddot_B2s()
 
-    \param A 2nd order tensor
-    \param B 2nd order tensor
-    \return scalar
-    */
-    template <class T>
-    inline T A2s_ddot_B2s(const T* A, const T* B);
+\param A 2nd order tensor
+\param B 2nd order tensor
+\return scalar
+*/
+template <class T>
+inline T A2s_ddot_B2s(const T* A, const T* B);
 
-    /**
-    See Cartesian2d::A2_dyadic_B2()
+/**
+See Cartesian2d::A2_dyadic_B2()
 
-    \param A 2nd order tensor
-    \param B 2nd order tensor
-    \param ret output 4th order tensor
-    */
-    template <class T>
-    inline void A2_dyadic_B2(const T* A, const T* B, T* ret);
+\param A 2nd order tensor
+\param B 2nd order tensor
+\param ret output 4th order tensor
+*/
+template <class T>
+inline void A2_dyadic_B2(const T* A, const T* B, T* ret);
 
-    /**
-    See Cartesian2d::A2_dot_B2()
+/**
+See Cartesian2d::A2_dot_B2()
 
-    \param A 2nd order tensor
-    \param B 2nd order tensor
-    \param ret output 2th order tensor
-    */
-    template <class T>
-    inline void A2_dot_B2(const T* A, const T* B, T* ret);
+\param A 2nd order tensor
+\param B 2nd order tensor
+\param ret output 2th order tensor
+*/
+template <class T>
+inline void A2_dot_B2(const T* A, const T* B, T* ret);
 
-    /**
-    See Cartesian2d::A4_ddot_B2()
+/**
+See Cartesian2d::A4_ddot_B2()
 
-    \param A 4th order tensor
-    \param B 2nd order tensor
-    \param ret output 2th order tensor
-    */
-    template <class T>
-    inline void A4_ddot_B2(const T* A, const T* B, T* ret);
+\param A 4th order tensor
+\param B 2nd order tensor
+\param ret output 2th order tensor
+*/
+template <class T>
+inline void A4_ddot_B2(const T* A, const T* B, T* ret);
 
 } // namespace pointer
 
