@@ -14,6 +14,30 @@ Implementation of GMatTensor/Cartesian2d.h
 namespace GMatTensor {
 namespace Cartesian2d {
 
+template <class T>
+inline auto underlying_size_A2(const T& A)
+{
+    return detail::impl_A2<T, 2>::toSizeT0(A.shape());
+}
+
+template <class T>
+inline auto underlying_size_A4(const T& A)
+{
+    return detail::impl_A4<T, 2>::toSizeT0(A.shape());
+}
+
+template <class T>
+inline auto underlying_shape_A2(const T& A)
+{
+    return detail::impl_A2<T, 2>::toShapeT0(A.shape());
+}
+
+template <class T>
+inline auto underlying_shape_A4(const T& A)
+{
+    return detail::impl_A4<T, 2>::toShapeT0(A.shape());
+}
+
 template <size_t N>
 inline Array<N>::Array(const std::array<size_t, N>& shape)
 {
@@ -32,6 +56,28 @@ inline void Array<N>::init(const std::array<size_t, N>& shape)
     m_size = std::accumulate(m_shape.begin(), m_shape.end(), 1, std::multiplies<size_t>());
 }
 
+template <size_t N>
+template <class T>
+inline auto Array<N>::view_tensor2(T& var, size_t i) const
+{
+    return xt::adapt(
+        &var.flat(i * m_stride_tensor2),
+        m_stride_tensor2,
+        xt::no_ownership(),
+        std::array<size_t, 2>{m_ndim, m_ndim});
+}
+
+template <size_t N>
+template <class T>
+inline auto Array<N>::view_tensor4(T& var, size_t i) const
+{
+    return xt::adapt(
+        &var.flat(i * m_stride_tensor4),
+        m_stride_tensor4,
+        xt::no_ownership(),
+        std::array<size_t, 4>{m_ndim, m_ndim, m_ndim, m_ndim});
+}
+
 template <size_t N>
 inline std::array<size_t, N> Array<N>::shape() const
 {
@@ -43,9 +89,9 @@ inline xt::xtensor<double, N + 2> Array<N>::O2() const
 {
     xt::xtensor<double, N + 2> ret = xt::empty<double>(m_shape_tensor2);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::O2(&ret.data()[i * m_stride_tensor2]);
+        Cartesian2d::pointer::O2(&ret.flat(i * m_stride_tensor2));
     }
 
     return ret;
@@ -56,9 +102,9 @@ inline xt::xtensor<double, N + 4> Array<N>::O4() const
 {
     xt::xtensor<double, N + 4> ret = xt::empty<double>(m_shape_tensor4);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::O4(&ret.data()[i * m_stride_tensor4]);
+        Cartesian2d::pointer::O4(&ret.flat(i * m_stride_tensor4));
     }
 
     return ret;
@@ -69,9 +115,9 @@ inline xt::xtensor<double, N + 2> Array<N>::I2() const
 {
     xt::xtensor<double, N + 2> ret = xt::empty<double>(m_shape_tensor2);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::I2(&ret.data()[i * m_stride_tensor2]);
+        Cartesian2d::pointer::I2(&ret.flat(i * m_stride_tensor2));
     }
 
     return ret;
@@ -82,9 +128,9 @@ inline xt::xtensor<double, N + 4> Array<N>::II() const
 {
     xt::xtensor<double, N + 4> ret = xt::empty<double>(m_shape_tensor4);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::II(&ret.data()[i * m_stride_tensor4]);
+        Cartesian2d::pointer::II(&ret.flat(i * m_stride_tensor4));
     }
 
     return ret;
@@ -95,9 +141,9 @@ inline xt::xtensor<double, N + 4> Array<N>::I4() const
 {
     xt::xtensor<double, N + 4> ret = xt::empty<double>(m_shape_tensor4);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::I4(&ret.data()[i * m_stride_tensor4]);
+        Cartesian2d::pointer::I4(&ret.flat(i * m_stride_tensor4));
     }
 
     return ret;
@@ -108,9 +154,9 @@ inline xt::xtensor<double, N + 4> Array<N>::I4rt() const
 {
     xt::xtensor<double, N + 4> ret = xt::empty<double>(m_shape_tensor4);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::I4rt(&ret.data()[i * m_stride_tensor4]);
+        Cartesian2d::pointer::I4rt(&ret.flat(i * m_stride_tensor4));
     }
 
     return ret;
@@ -121,9 +167,9 @@ inline xt::xtensor<double, N + 4> Array<N>::I4s() const
 {
     xt::xtensor<double, N + 4> ret = xt::empty<double>(m_shape_tensor4);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::I4s(&ret.data()[i * m_stride_tensor4]);
+        Cartesian2d::pointer::I4s(&ret.flat(i * m_stride_tensor4));
     }
 
     return ret;
@@ -134,9 +180,9 @@ inline xt::xtensor<double, N + 4> Array<N>::I4d() const
 {
     xt::xtensor<double, N + 4> ret = xt::empty<double>(m_shape_tensor4);
 
-    #pragma omp parallel for
+#pragma omp parallel for
     for (size_t i = 0; i < m_size; ++i) {
-        Cartesian2d::pointer::I4d(&ret.data()[i * m_stride_tensor4]);
+        Cartesian2d::pointer::I4d(&ret.flat(i * m_stride_tensor4));
     }
 
     return ret;
